@@ -6,14 +6,16 @@ extends 'Zonemaster::Engine::Exception';
 # The actual interesting module.
 package Zonemaster::CLI;
 
-use version; our $VERSION = version->declare("v2.0.4");
-
 use 5.014002;
+
+use strict;
 use warnings;
 
+use version; our $VERSION = version->declare( "v3.2.0" );
+
 use Locale::TextDomain 'Zonemaster-CLI';
 use Moose;
-with 'MooseX::Getopt';
+with 'MooseX::Getopt::GLD' => { getopt_conf => [ 'pass_through' ] };
 
 use Zonemaster::Engine;
 use Zonemaster::Engine::Logger::Entry;
@@ -21,6 +23,7 @@ use Zonemaster::Engine::Translator;
 use Zonemaster::Engine::Util qw[pod_extract_for];
 use Zonemaster::Engine::Exception;
 use Zonemaster::Engine::Zone;
+use Zonemaster::Engine::Net::IP;
 use Scalar::Util qw[blessed];
 use Encode;
 use Zonemaster::LDNS;
@@ -29,6 +32,7 @@ use List::Util qw[max];
 use Text::Reflow qw[reflow_string];
 use JSON::XS;
 use File::Slurp;
+use Socket qw[AF_INET AF_INET6];
 
 our %numeric = Zonemaster::Engine::Logger::Entry->levels;
 our $JSON    = JSON::XS->new->allow_blessed->convert_blessed->canonical;
@@ -44,11 +48,11 @@ has 'version' => (
 );
 
 has 'level' => (
-    is          => 'ro',
-    isa         => 'Str',
-    required    => 0,
-    default     => 'NOTICE',
-    initializer => sub {
+    is            => 'ro',
+    isa           => 'Str',
+    required      => 0,
+    default       => 'NOTICE',
+    initializer   => sub {
         my ( $self, $value, $set, $attr ) = @_;
         $set->( uc $value );
     },
@@ -71,16 +75,22 @@ has 'json' => (
 );
 
 has 'json_stream' => (
+    traits        => [ 'Getopt' ],
     is            => 'ro',
     isa           => 'Bool',
     default       => 0,
+    cmd_aliases   => 'json_stream',
+    cmd_flag      => 'json-stream',
     documentation => __( 'Flag indicating if output should be streaming JSON or not.' ),
 );
 
 has 'json_translate' => (
+    traits        => [ 'Getopt' ],
     is            => 'ro',
     isa           => 'Bool',
     default       => 0,
+    cmd_aliases   => 'json_translate',
+    cmd_flag      => 'json-translate',
     documentation => __( 'Flag indicating if streaming JSON output should include the translated message of the tag or not.' ),
 );
 
@@ -99,19 +109,35 @@ has 'time' => (
 );
 
 has 'show_level' => (
+    traits        => [ 'Getopt' ],
     is            => 'ro',
     isa           => 'Bool',
+    cmd_aliases   => 'show_level',
+    cmd_flag      => 'show-level',
     documentation => __( 'Print level on entries.' ),
     default       => 1,
 );
 
 has 'show_module' => (
+    traits        => [ 'Getopt' ],
     is            => 'ro',
     isa           => 'Bool',
+    cmd_aliases   => 'show_module',
+    cmd_flag      => 'show-module',
     documentation => __( 'Print the name of the module on entries.' ),
     default       => 0,
 );
 
+has 'show_testcase' => (
+    traits        => [ 'Getopt' ],
+    is            => 'ro',
+    isa           => 'Bool',
+    cmd_aliases   => 'show_testcase',
+    cmd_flag      => 'show-testcase',
+    documentation => __( 'Print the name of the test case on entries.' ),
+    default       => 0,
+);
+
 has 'ns' => (
     is            => 'ro',
     isa           => 'ArrayRef',
@@ -133,25 +159,26 @@ has 'restore' => (
 );
 
 has 'ipv4' => (
-    is      => 'ro',
-    isa     => 'Bool',
-    default => 1,
+    is            => 'ro',
+    isa           => 'Bool',
     documentation =>
       __( 'Flag to permit or deny queries being sent via IPv4. --ipv4 permits IPv4 traffic, --no-ipv4 forbids it.' ),
 );
 
 has 'ipv6' => (
-    is      => 'ro',
-    isa     => 'Bool',
-    default => 1,
+    is            => 'ro',
+    isa           => 'Bool',
     documentation =>
       __( 'Flag to permit or deny queries being sent via IPv6. --ipv6 permits IPv6 traffic, --no-ipv6 forbids it.' ),
 );
 
 has 'list_tests' => (
+    traits        => [ 'Getopt' ],
     is            => 'ro',
     isa           => 'Bool',
     default       => 0,
+    cmd_aliases   => 'list_tests',
+    cmd_flag      => 'list-tests',
     documentation => __( 'Instead of running a test, list all available tests.' ),
 );
 
@@ -165,13 +192,16 @@ has 'test' => (
 );
 
 has 'stop_level' => (
-    is          => 'ro',
-    isa         => 'Str',
-    required    => 0,
-    initializer => sub {
+    traits        => [ 'Getopt' ],
+    is            => 'ro',
+    isa           => 'Str',
+    required      => 0,
+    initializer   => sub {
         my ( $self, $value, $set, $attr ) = @_;
         $set->( uc $value );
     },
+    cmd_aliases   => 'stop_level',
+    cmd_flag      => 'stop-level',
     documentation => __(
 'As soon as a message at this level or higher is logged, execution will stop. Must be one of CRITICAL, ERROR, WARNING, NOTICE, INFO or DEBUG.'
     )
@@ -184,20 +214,6 @@ has 'profile' => (
     documentation => __( 'Name of profile file to load. (DEFAULT)' ),
 );
 
-has 'config' => (
-    is            => 'ro',
-    isa           => 'Str',
-    required      => 0,
-    documentation => __( 'Name of configuration file to load. (TERMINATED)' ),
-);
-
-has 'policy' => (
-    is            => 'ro',
-    isa           => 'Str',
-    required      => 0,
-    documentation => __( 'Name of policy file to load. (TERMINATED)' ),
-);
-
 has 'ds' => (
     is            => 'ro',
     isa           => 'ArrayRef[Str]',
@@ -216,13 +232,13 @@ has 'progress' => (
     is            => 'ro',
     isa           => 'Bool',
     default       => !!( -t STDOUT ),
-    documentation => __( 'Boolean flag for activity indicator. Defaults to on if STDOUT is a tty, off if it is not. Disable with --noprogress.' ),
+    documentation => __( 'Boolean flag for activity indicator. Defaults to on if STDOUT is a tty, off if it is not. Disable with --no-progress.' ),
 );
 
 has 'encoding' => (
-    is      => 'ro',
-    isa     => 'Str',
-    default => sub {
+    is            => 'ro',
+    isa           => 'Str',
+    default       => sub {
         my $locale = $ENV{LC_CTYPE} // 'C';
         my ( $e ) = $locale =~ m|\.(.*)$|;
         $e //= 'UTF-8';
@@ -240,41 +256,31 @@ has 'nstimes' => (
 );
 
 has 'dump_profile' => (
-    is => 'ro',
-    isa => 'Bool',
-    required => 0,
-    default => 0,
+    traits        => [ 'Getopt' ],
+    is            => 'ro',
+    isa           => 'Bool',
+    required      => 0,
+    default       => 0,
+    cmd_aliases   => 'dump_profile',
+    cmd_flag      => 'dump-profile',
     documentation => __( 'Print the effective profile used in JSON format, then exit.' ),
 );
 
-has 'dump_config' => (
-    is => 'ro',
-    isa => 'Bool',
-    required => 0,
-    default => 0,
-    documentation => __( 'Print the effective configuration used in JSON format, then exit. (TERMINATED)' ),
-);
-
-has 'dump_policy' => (
-    is => 'ro',
-    isa => 'Bool',
-    required => 0,
-    default => 0,
-    documentation => __( 'Print the effective policy used in JSON format, then exit. (TERMINATED)' ),
-);
-
 has 'sourceaddr' => (
-    is => 'ro',
-    isa => 'Str',
-    required => 0,
-    documentation => __( 'Local IP address that the test engine should try to send its requests from.' ),
+    is            => 'ro',
+    isa           => 'Str',
+    required      => 0,
+    documentation => __(
+            'Source IP address used to send queries. '
+          . 'Setting an IP address not correctly configured on a local network interface causes cryptic error messages.'
+    ),
 );
 
 has 'elapsed' => (
-    is => 'ro',
-    isa => 'Bool',
-    required => 0,
-    default => 0,
+    is            => 'ro',
+    isa           => 'Bool',
+    required      => 0,
+    default       => 0,
     documentation => 'Print elapsed time at end of run.',
 );
 
@@ -284,6 +290,12 @@ sub run {
     my %counter;
     my $printed_something;
 
+    if ( grep /^-/, @{ $self->extra_argv } ) {
+        print "Unknown option: ", join( q{ }, grep /^-/, @{ $self->extra_argv } ), "\n";
+        print "Run \"zonemaster-cli -h\" to get the valid options\n";
+        exit;
+    }
+
     if ( $self->locale ) {
         undef $ENV{LANGUAGE};
         $ENV{LC_ALL} = $self->locale;
@@ -308,52 +320,38 @@ sub run {
         print_test_list();
     }
 
-    Zonemaster::Engine::Profile->effective->set( q{net.ipv4}, 0+$self->ipv4 );
-    Zonemaster::Engine::Profile->effective->set( q{net.ipv6}, 0+$self->ipv6 );
-
     if ($self->sourceaddr) {
+
         Zonemaster::Engine::Profile->effective->set( q{resolver.source}, $self->sourceaddr );
     }
 
     # Filehandle for diagnostics output
-    my $fh_diag = ( $self->json or $self->json_stream or $self->raw or $self->dump_config or $self->dump_policy )
+    my $fh_diag = ( $self->json or $self->json_stream or $self->raw )
       ? *STDERR     # Structured output mode (e.g. JSON)
       : *STDOUT;    # Human readable output mode
 
     if ( $self->profile ) {
         say $fh_diag __x( "Loading profile from {path}.", path => $self->profile );
-	my $json    = read_file( $self->profile );
-	my $foo     = Zonemaster::Engine::Profile->from_json( $json );
-	my $profile = Zonemaster::Engine::Profile->default;
-	$profile->merge( $foo );
-	Zonemaster::Engine::Profile->effective->merge( $profile );
+        my $json    = read_file( $self->profile );
+        my $foo     = Zonemaster::Engine::Profile->from_json( $json );
+        my $profile = Zonemaster::Engine::Profile->default;
+        $profile->merge( $foo );
+        Zonemaster::Engine::Profile->effective->merge( $profile );
     }
-    else {
 
-        if ( $self->policy ) {
-            say $fh_diag __x( "Loading policy from {path}.", path => $self->policy );
-            say $fh_diag __x( "Use of config and policy have been TERMINATED, use profile instead." );
-            exit( 1 );
-        }
-
-        if ( $self->config ) {
-            say $fh_diag __x( "Loading configuaration from {path}.", path => $self->config );
-            say $fh_diag __x( "Use of config and policy have been TERMINATED, use profile instead." );
-            exit( 1 );
-        }
+    # These two must come after any profile from command line has been loaded
+    # to make any IPv4/IPv6 option override the profile setting.
+    if ( defined ($self->ipv4) ) {
+        Zonemaster::Engine::Profile->effective->set( q{net.ipv4}, 0+$self->ipv4 );
     }
+    if ( defined ($self->ipv6) ) {
+        Zonemaster::Engine::Profile->effective->set( q{net.ipv6}, 0+$self->ipv6 );
+    }
+
 
     if ( $self->dump_profile ) {
         do_dump_profile();
     }
-    else {
-
-        if ( $self->dump_config or $self->dump_policy ) {
-            say $fh_diag __x( "TERMINATED, use dump_profile instead." );
-            exit( 1 );
-        }
-
-    }
 
     if ( $self->stop_level and not defined( $numeric{ $self->stop_level } ) ) {
         die __( "Failed to recognize stop level '" ) . $self->stop_level . "'.\n";
@@ -366,13 +364,11 @@ sub run {
     my $translator;
     $translator = Zonemaster::Engine::Translator->new unless ( $self->raw or $self->json or $self->json_stream );
     $translator->locale( $self->locale ) if $translator and $self->locale;
-    eval { $translator->data } if $translator;    # Provoke lazy loading of translation data
 
     my $json_translator;
     if ( $self->json_translate ) {
         $json_translator = Zonemaster::Engine::Translator->new;
         $json_translator->locale( $self->locale ) if $self->locale;
-        eval { $json_translator->data };
     }
 
     if ( $self->restore ) {
@@ -392,25 +388,44 @@ sub run {
                 $printed_something = 1;
 
                 if ( $translator ) {
+                    my $header = q{};
                     if ( $self->time ) {
-                        printf "%7.2f ", $entry->timestamp;
+                        $header .= sprintf "%7.2f ", $entry->timestamp;
                     }
 
                     if ( $self->show_level ) {
-                        printf "%-9s ", __( $entry->level );
+                        $header .= sprintf "%-9s ", translate_severity( $entry->level );
                     }
 
                     if ( $self->show_module ) {
-                        printf "%-12s ", $entry->module;
+                        $header .= sprintf "%-12s ", $entry->module;
                     }
 
-                    say $translator->translate_tag( $entry );
+                    if ( $self->show_testcase ) {
+                        $header .= sprintf "%-14s ", $entry->testcase;
+                    }
+
+                    print $header;
+
+                    if ( $entry->level eq q{DEBUG3} and scalar( keys %{$entry->args} ) == 1 and defined $entry->args->{packet} ) {
+                        my $packet = $entry->args->{packet};
+                        my $padding = q{ } x length $header;
+                        $entry->args->{packet} = q{};
+                        say $translator->translate_tag( $entry );
+                        foreach my $line ( split /\n/, $packet ) {
+                            print $padding, $line, "\n";
+                        }
+                    }
+                    else {
+                        say $translator->translate_tag( $entry );
+                    }
                 }
                 elsif ( $self->json_stream ) {
                     my %r;
 
                     $r{timestamp} = $entry->timestamp;
                     $r{module}    = $entry->module;
+                    $r{testcase}  = $entry->testcase;
                     $r{tag}       = $entry->tag;
                     $r{level}     = $entry->level;
                     $r{args}      = $entry->args if $entry->args;
@@ -421,11 +436,24 @@ sub run {
                 elsif ( $self->json ) {
                     # Don't do anything
                 }
-                elsif ( $self->show_module ) {
-                    printf "%7.2f %-9s %-12s %s\n", $entry->timestamp, $entry->level, $entry->module, $entry->string;
-                }
                 else {
-                    printf "%7.2f %-9s %s\n", $entry->timestamp, $entry->level, $entry->string;
+                    my $prefix = sprintf "%7.2f %-9s ", $entry->timestamp, $entry->level;
+                    if ( $self->show_module ) {
+                        $prefix .= sprintf "%-12s ", $entry->module;
+                    }
+                    if ( $self->show_testcase ) {
+                        $prefix .= sprintf "%-14s ", $entry->testcase;
+                    }
+                    $prefix .= $entry->tag;
+
+                    my $message = $entry->string;
+                    $message =~ s/^[A-Z0-9:_]+//;    # strip MODULE:TAG, they're coming in $prefix instead
+                    my @lines = split /\n/, $message;
+
+                    printf "%s%s %s\n", $prefix, ' ', shift @lines;
+                    for my $line ( @lines ) {
+                        printf "%s%s %s\n", $prefix, '>', $line;
+                    }
                 }
             } ## end if ( $numeric{ uc $entry...})
             if ( $self->stop_level and $numeric{ uc $entry->level } >= $numeric{ $self->stop_level } ) {
@@ -434,7 +462,7 @@ sub run {
         }
     );
 
-    if ( $self->profile or $self->config or $self->policy ) {
+    if ( $self->profile ) {
         # Separate initialization from main output in human readable output mode
         print "\n" if $fh_diag eq *STDOUT;
     }
@@ -454,6 +482,9 @@ sub run {
         if ( $self->show_module ) {
             print __( 'Module       ' );
         }
+        if ( $self->show_testcase ) {
+            print __( 'Testcase       ' );
+        }
         say __( 'Message' );
 
         if ( $self->time ) {
@@ -465,6 +496,9 @@ sub run {
         if ( $self->show_module ) {
             print __( '============ ' );
         }
+        if ( $self->show_testcase ) {
+            print __( '============== ' );
+        }
         say __( '=======' );
     } ## end if ( $translator )
 
@@ -515,13 +549,13 @@ sub run {
         say __( "\n\n   Level\tNumber of log entries" );
         say "   =====\t=====================";
         foreach my $level ( sort { $numeric{$b} <=> $numeric{$a} } keys %counter ) {
-            printf __( "%8s\t%5d entries.\n" ), __( $level ), $counter{$level};
+            printf __( "%8s\t%5d entries.\n" ), translate_severity( $level ), $counter{$level};
         }
     }
 
     if ( $self->nstimes ) {
         my $zone = Zonemaster::Engine->zone( $domain );
-        my $max = max map { length( "$_" ) } @{ $zone->ns };
+        my $max = max map { length( "$_" ) } ( @{ $zone->ns }, q{Server} );
 
         print "\n";
         printf "%${max}s %s\n", 'Server', ' Max (ms)      Min      Avg   Stddev   Median     Total';
@@ -676,6 +710,31 @@ sub do_dump_profile {
     exit;
 }
 
+sub translate_severity {
+    my $severity = shift;
+    if ( $severity eq "DEBUG" ) {
+        return __( "DEBUG" );
+    }
+    elsif ( $severity eq "INFO" ) {
+        return __( "INFO" );
+    }
+    elsif ( $severity eq "NOTICE" ) {
+        return __( "NOTICE" );
+    }
+    elsif ( $severity eq "WARNING" ) {
+        return __( "WARNING" );
+    }
+    elsif ( $severity eq "ERROR" ) {
+        return __( "ERROR" );
+    }
+    elsif ( $severity eq "CRITICAL" ) {
+        return __( "CRITICAL" );
+    }
+    else {
+        return $severity;
+    }
+}
+
 1;
 
 __END__
